@@ -1,6 +1,7 @@
 /* test-case-helper.c
  *
  * Copyright © 2013 Intel Corporation
+ * Copyright © 2015 Collabora Ltd.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -29,6 +30,8 @@
 typedef struct {
     gpointer self;
     FolksTestCaseTestMethod test;
+    gpointer test_data;
+    GDestroyNotify test_data_free;
 } FolksTestCaseWeakMethod;
 
 static void
@@ -52,7 +55,7 @@ folks_test_case_weak_method_test (gpointer fixture G_GNUC_UNUSED,
   g_return_if_fail (wm->self != NULL);
   g_return_if_fail (FOLKS_IS_TEST_CASE (wm->self));
 
-  wm->test (wm->self);
+  wm->test (wm->test_data);
 }
 
 static void
@@ -67,21 +70,33 @@ folks_test_case_weak_method_teardown (gpointer fixture G_GNUC_UNUSED,
   folks_test_case_tear_down (wm->self);
 }
 
+static void
+test_case_destroyed_cb (gpointer data,
+    GObject *test_case_location)
+{
+  FolksTestCaseWeakMethod *wm = data;
+
+  if (wm->test_data != NULL && wm->test_data_free != NULL)
+    wm->test_data_free (wm->test_data);
+}
+
 GTestCase *
 folks_test_case_add_test_helper (FolksTestCase *self,
     const gchar *name,
     FolksTestCaseTestMethod test,
-    void *test_target)
+    void *test_target,
+    GDestroyNotify test_target_destroy_notify)
 {
   FolksTestCaseWeakMethod *wm;
 
-  g_return_val_if_fail (self == (FolksTestCase *) test_target, NULL);
-
   /* This will never be freed, so make sure not to hold references. */
   wm = g_new0 (FolksTestCaseWeakMethod, 1);
   wm->self = self;
   wm->test = test;
-  g_object_add_weak_pointer (G_OBJECT (self), &wm->self);
+  wm->test_data = test_target;
+  wm->test_data_free = test_target_destroy_notify;
+
+  g_object_weak_ref (G_OBJECT (self), test_case_destroyed_cb, wm);
 
   return g_test_create_case (name,
       0,
